//
// Copyright (c) 2008 smdn <smdn@smdn.jp>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
using System;
using System.Collections.Generic;
using System.Text;

using Smdn.Text;

namespace Smdn.Formats.Mime.Formatting {
  public class MimeHeaderBuilder : IHeaderFieldBody {
    internal const string NameBodyFormattingDelimiter = ": ";

    public static MimeHeaderBuilder CreateContentType(MimeType mimeType)
    {
      return CreateContentType(null, mimeType, null, null, null);
    }

    public static MimeHeaderBuilder CreateContentType(MimeType mimeType, Encoding charset)
    {
      return CreateContentType(null, mimeType, charset, null, null);
    }

    public static MimeHeaderBuilder CreateContentType(MimeType mimeType, Encoding charset, string boundary)
    {
      return CreateContentType(null, mimeType, charset, boundary, null);
    }

    public static MimeHeaderBuilder CreateContentType(MimeType mimeType, Encoding charset, string boundary, string name)
    {
      return CreateContentType(null, mimeType, charset, boundary, name);
    }

    public static MimeHeaderBuilder CreateContentType(MimeFormat format, MimeType mimeType, Encoding charset)
    {
      return CreateContentType(format, mimeType, charset, null, null);
    }

    public static MimeHeaderBuilder CreateContentType(MimeFormat format, MimeType mimeType, Encoding charset, string boundary)
    {
      return CreateContentType(format, mimeType, charset, boundary, null);
    }

    public static MimeHeaderBuilder CreateContentType(MimeFormat format, MimeType mimeType, Encoding charset, string boundary, string name)
    {
      if (mimeType == null)
        throw new ArgumentNullException(nameof(mimeType));

      var fragments = new List<MimeHeaderFragment>(4);

      fragments.Add(new MimeHeaderFragment(mimeType.ToString()));

      if (charset != null) {
        var charsetName =
#if NETFRAMEWORK || NETSTANDARD2_0
          charset.BodyName;
#else
          charset.WebName;
#endif

        fragments.Add(new MimeHeaderParameterFragment("charset", charsetName));
      }

      if (boundary != null)
        fragments.Add(new MimeHeaderParameterFragment("boundary", boundary));

      if (name != null)
        // TODO: name*n*
        fragments.Add(new MimeHeaderParameterFragment("name", MimeEncoding.Encode(name, MimeEncodingMethod.Base64, Charsets.UTF8NoBom)));

      return new MimeHeaderBuilder(MimeHeaderNames.ContentType, format, fragments);
    }

    public static MimeHeaderBuilder CreateContentTransferEncoding(ContentTransferEncodingMethod transferEncoding)
    {
      return CreateContentTransferEncoding(null, transferEncoding);
    }

    public static MimeHeaderBuilder CreateContentTransferEncoding(MimeFormat format, ContentTransferEncodingMethod transferEncoding)
    {
      return new MimeHeaderBuilder(ContentTransferEncoding.HeaderName,
                                   format,
                                   new[] {new MimeHeaderFragment(ContentTransferEncoding.GetEncodingName(transferEncoding))});
    }

    public static MimeHeaderBuilder CreateContentDisposition(MimeMessageDisposition disposition, string filename)
    {
      var fragments = new List<MimeHeaderFragment>(2);

      switch (disposition) {
        case MimeMessageDisposition.Inline: fragments.Add(new MimeHeaderFragment("inline")); break;
        case MimeMessageDisposition.Attachment: fragments.Add(new MimeHeaderFragment("attachment")); break;
        default:
          throw ExceptionUtils.CreateArgumentMustBeValidEnumValue(nameof(disposition), disposition, "must be inline or attachment");
      }

      if (!string.IsNullOrEmpty(filename))
        // TODO: filename*n*
        fragments.Add(new MimeHeaderParameterFragment("filename", MimeEncoding.Encode(filename, MimeEncodingMethod.Base64, Charsets.UTF8NoBom)));

      return new MimeHeaderBuilder(MimeHeaderNames.ContentDisposition, fragments);
    }

    /*
     * instance members
     */
    public string Name {
      get { return name; }
    }

    public MimeFormat Format {
      get; set;
    }

    public string Value {
      get; set;
    }

    public MimeEncodingMethod Encoding {
      get; set;
    }

    public Encoding Charset {
      get; set;
    }

    public MimeHeaderBuilder(string name)
      : this(name, null, MimeFormat.Standard, MimeEncodingMethod.None, null)
    {
    }

    public MimeHeaderBuilder(string name, string @value)
      : this(name, @value, MimeFormat.Standard, MimeEncodingMethod.None, null)
    {
    }

    public MimeHeaderBuilder(string name, string @value, MimeFormat format)
      : this(name, @value, format, MimeEncodingMethod.None, null)
    {
    }

    public MimeHeaderBuilder(string name, string @value, MimeEncodingMethod encoding)
      : this(name, @value, MimeFormat.Standard, encoding, null)
    {
    }

    public MimeHeaderBuilder(string name, string @value, MimeEncodingMethod encoding, Encoding charset)
      : this(name, @value, MimeFormat.Standard, encoding, charset)
    {
    }

    public MimeHeaderBuilder(string name, string @value, MimeFormat format, MimeEncodingMethod encoding)
      : this(name, @value, format, encoding, null)
    {
    }

    public MimeHeaderBuilder(string name, IEnumerable<MimeHeaderFragment> headerFragments)
      : this(name, string.Empty, MimeFormat.Standard, MimeEncodingMethod.None, null)
    {
      SetValue(null, headerFragments);
    }

    public MimeHeaderBuilder(string name, MimeFormat format, IEnumerable<MimeHeaderFragment> headerFragments)
      : this(name, string.Empty, format, MimeEncodingMethod.None, null)
    {
      SetValue(format, headerFragments);
    }

    public MimeHeaderBuilder(string name, string @value, MimeFormat format, MimeEncodingMethod encoding, Encoding charset)
    {
      if (name == null)
        throw new ArgumentNullException(nameof(name));
      if (name.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(nameof(name));

      // TODO: check name
      // field-name  =  1*<any CHAR, excluding CTLs, SPACE, and ":">
      this.name = name;
      this.Format = format;
      this.Value = @value;
      this.Encoding = encoding;
      this.Charset = charset;
    }

    public void SetValue(MimeFormat format, IEnumerable<MimeHeaderFragment> headerFragments)
    {
      if (headerFragments == null)
        throw new ArgumentNullException(nameof(headerFragments));

      Format = format;
      Encoding = MimeEncodingMethod.None;
      Charset  = null;
      Value = EncodeFragments(headerFragments);
    }

    private string EncodeFragments(IEnumerable<MimeHeaderFragment> fragments)
    {
      var sb = new StringBuilder();
      var first = true;
      var currentLineLength = name.Length + NameBodyFormattingDelimiter.Length;
      var encodedSplitter = new[] {Ascii.Chars.CRLF + Ascii.Chars.HT};
      var format = Format ?? MimeFormat.Unspecified;

      foreach (var fragment in fragments) {
        var encodedLines = Encoder.Encode(fragment, format).Split(encodedSplitter, StringSplitOptions.None);

        for (var line = 0; line < encodedLines.Length; line++) {
          // delimiter
          if (line == 0 && fragment is MimeHeaderParameterFragment) {
            sb.Append(';');
            currentLineLength += 1;
          }

          // folding
          if (0 < format.Folding && format.Folding <= currentLineLength + encodedLines[line].Length) {
            sb.Append(format.GetHeaderFoldingString());
            sb.Append(encodedLines[line]);

            currentLineLength = 1/*TAB*/ + encodedLines[line].Length;
          }
          else {
            if (!first && line == 0) {
              // delimiter
              sb.Append(Ascii.Chars.SP);
              currentLineLength = 1;
            }

            sb.Append(encodedLines[line]);

            currentLineLength += encodedLines[line].Length;
          }
        }

        first = false;
      } // foreach fragment

      return sb.ToString();
    }

    public bool IsNameEquals(string name)
    {
      return string.Equals(this.name, name, StringComparison.OrdinalIgnoreCase);
    }

    public override string ToString()
    {
      return string.Concat(name,
                           MimeHeaderBuilder.NameBodyFormattingDelimiter,
                           Value);
    }

    private string name;
  }
}

